import inspect

from voluptuous import Schema

from yesaide.database import MetaBase


def update(sqla_obj, schema, ignore_keys=None, **kwargs):
    """Update an `instance`. Return False if there is no update and
    True otherwise.

    Do not raise error if too many arguments are given.

    Do not commit the database session.

    Keyword arguments:
        sqla_obj -- SQLAlchemy object to update
        schema -- voluptuous schema to perform data validation
        ignore_keys -- list of keys that will be ignored by this
                       function

    """
    if not isinstance(schema, Schema):
        raise AttributeError('`schema` must be a voluptuous schema.')

    if ignore_keys is None:
        ignore_keys = []

    # Explicitely cast to string properties which come from schema
    # to deal with `voluptuous.Required` stuff.
    schema_keys = set([str(k) for k in schema.schema if k not in ignore_keys])

    obj_current_dict = {
        k: getattr(sqla_obj, k) for k in schema_keys
        if not getattr(sqla_obj, k) is None
    }
    obj_update_dict = obj_current_dict.copy()

    to_update = schema_keys.intersection(kwargs.keys())

    for item in to_update:
        obj_update_dict[item] = kwargs[item]

    obj_update_dict = schema(obj_update_dict)

    for item in to_update:
        setattr(sqla_obj, item, obj_update_dict[item])

    return obj_update_dict != obj_current_dict


def resolve_id(build_query, a_dict, schema, allow_none_id=False):
    """Return a dict fulfilled with the missing objects according to
    the given dict (in `a_dict`) and `schema`.

    Will fetch `sqla_obj` if `sqla_obj_id` (or `sqla_obj_uuid`) is in
    the dict keys, if `sqla_obj` is accepted by the schema and is
    supposed to be a SQLAlchemy object.

    Additionnally if `sqla_obj_id` is not in the schema, it will be
    removed from the dict.

    If `allow_none_id` is true, passing `sqla_obj_id` with value
    'None' or '""' (empty string) will result in setting `sqla_obj`
    to 'None'.

    Arguments:
        build_query -- function which will be given a SQLAlchemy mapping
                       and must return an appropriate query
        a_dict -- base dictionnary
        schema -- reference voluptuous schema
        allow_none_id -- bool (see above)

    """
    _a_dict = a_dict.copy()

    for key, sqla_map in schema.schema.items():
        if key in _a_dict.keys():
            continue

        if not inspect.isclass(sqla_map) or not issubclass(sqla_map, MetaBase):
            continue

        key_id = '{}_id'.format(key)
        key_uuid = '{}_uuid'.format(key)

        if key_id in a_dict and key_uuid in a_dict:
            raise Exception('`_resolve_id()` has been called with '
                            'both an `object_id` and an '
                            '`object_uuid`.')

        elif key_id in a_dict:
            val_id = _a_dict.pop(key_id)
            if not val_id and allow_none_id:
                val = None

            else:
                val = build_query(sqla_map).filter(sqla_map.id == val_id).one()

            _a_dict[str(key)] = val

        elif key_uuid in a_dict:
            val_uuid = _a_dict.pop(key_uuid)
            if not val_uuid and allow_none_id:
                val = None

            else:
                val = build_query(sqla_map)\
                    .filter(sqla_map.uuid == val_uuid).one()

            _a_dict[str(key)] = val

    return _a_dict



import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.types import TypeDecorator, CHAR


class MetaBase(object):
    """A meta base class for declarative base.

    Example usage:

        isinstance(SomeSQLAClass, MetaBase):
            ...

    """


def db_method(func):
    """Decorator for a database method of a `DataRepository` object.

    The decorated method's object must have its database session
    stored in self._dbsession`.

    """
    def wrapped_commit_func(self, *args, **kwargs):
        # Determine if we'll issue a commit or not. Remove 'commit'
        # from kwargs anyway.
        commit = kwargs.pop('commit', True)
        if getattr(self._dbsession, '_ya_in_transaction', False):
            commit = False

        retval = func(self, *args, **kwargs)

        if commit:
            self._dbsession.commit()

        return retval

    return wrapped_commit_func


class GUIDType(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses CHAR(32), storing as
    stringified hex values.

    Inspired from: http://docs.sqlalchemy.org/en/rel_0_8/core/types.html

    """

    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is not None:
            return uuid.UUID(value)
        return None



import imp
import os


class ConfigError(KeyError):
    """Exception thrown when the requested key does not exist."""


class Config(object):
    """Has a dict-like interface with some handy subtilities regarding
    config management.

    """

    def __init__(self, env_prefix=None):
        self.env_prefix = env_prefix

        self.base_values = {}
        self.set_values = {}

    def __getitem__(self, name):
        try:
            return self.set_values[name]
        except KeyError:
            pass

        if self.env_prefix:
            try:
                return os.environ[self.env_prefix+name]
            except KeyError:
                pass

        try:
            return self.base_values[name]
        except KeyError:
            raise ConfigError(
                'The requested config value, {}, is not set.'.format(name))

    def __setitem__(self, name, value):
        self.set_values[name] = value

    def get(self, name, default_value=None):
        try:
            return self.__getitem__(name)
        except ConfigError:
            return default_value

    def from_object(self, obj):
        for key in dir(obj):
            if key.isupper():
                self.base_values[key] = getattr(obj, key)

    def from_pyfile(self, filename):
        d = imp.new_module('config')
        d.__file__ = filename
        with open(filename, mode='rb') as config_file:
            exec(compile(config_file.read(), filename, 'exec'), d.__dict__)

        self.from_object(d)



import re


def set_filename_version(filename, version_number, pattern):
    """Write the given version number in the given (in filename) file,
    after the given pattern.

    E.g.:
        set_filename_version('setup.py', '1.0.0', 'version')

    Extracted from Flask source, see:
        - file: flask/scripts/make-release.py
        - commit: 54688b26a9562c9a9c31541dbf32e0d8ee7a5926

    """
    def inject_version(match):
        before, old, after = match.groups()
        return before + version_number + after

    with open(filename) as f:
        contents = re.sub(r"^(\s*%s\s*=\s*')(.+?)(')(?sm)" % pattern,
                          inject_version, f.read())

    with open(filename, 'w') as f:
        f.write(contents)


def compute_semver(current_version, release='normal', compute_dev=True):
    """Compute and return the next version number, given a the current
    version and the release type ('normal', 'minor' or 'major'), as well
    as the next dev version (if `compute_dev` is 'True').

    """
    if current_version.endswith('dev'):
        dev_version = current_version.replace('dev', '')
        dev_version_parts = list(map(int, dev_version.split('.')))
        current_version_parts = dev_version_parts
        current_version_parts[2] -= 1
    else:
        current_version_parts = list(map(int, current_version.split('.')))

    release_version_parts = current_version_parts
    if release == 'minor':
        release_version_parts[2] += 1
    elif release == 'normal':
        release_version_parts[1] += 1
        release_version_parts[2] = 0
    elif release == 'major':
        release_version_parts[0] += 1
        release_version_parts[1] = 0
        release_version_parts[2] = 0

    release_version = '.'.join(map(str, release_version_parts))

    next_dev_version = None
    if compute_dev:
        next_dev_version, _ = compute_semver(
            release_version, release='minor', compute_dev=False)
        next_dev_version = '{}dev'.format(next_dev_version)

    return release_version, next_dev_version


def flags_to_release(is_minor=False, is_major=False):
    """Convert flags to release type."""
    if is_minor and is_major:
        raise ValueError("Both `is_minor` and `is_major` are set to 'True'.")

    if is_minor:
        return 'minor'
    if is_major:
        return 'major'
    return 'normal'



from yesaide.worker import RawWorker


class RawForeman(RawWorker):
    """Special worker to manage other workers."""

    def __init__(self, dbsession=None):
        RawWorker.__init__(self, dbsession)



"""Validation utilities and extra schemas. Mostly built around
voluptuous.

"""
import datetime
import decimal
import math
import re
import uuid

from voluptuous import Invalid, Required, Schema


def is_valid_uuid(value):
    if isinstance(value, uuid.UUID):
        return True

    try:
        maybe_value = uuid.UUID(value, version=4)
    except ValueError:
        return False

    return maybe_value.hex == value.replace('-', '').replace(' ', '')


_mail_regexp = re.compile('[^@]+@[^@]+\.[^@]+')


def is_valid_mail(raw_mail):
    if re.match(_mail_regexp, raw_mail) and ' ' not in raw_mail:
        return True
    return False


def Mail(empty_to_none=False, msg=None, lower=False):
    def f(value):
        if value in [None, ''] and empty_to_none:
            return None

        if not is_valid_mail(value):
            raise Invalid(msg or ('Incorrect mail address.'))

        if lower:
            return value.lower()
        return value
    return f


def Integeable(empty_to_none=False, cast=True, msg=None):
    def f(value):
        if value in [None, ''] and empty_to_none:
            return None

        try:
            casted_value = int(value)
        except (ValueError, TypeError):
            raise Invalid(msg or 'Given value cannot be casted to int.')

        if str(value) != str(casted_value):
            raise Invalid(msg or 'Given value cannot be casted to int.')

        if cast:
            return casted_value
        return value
    return f


def Floatable(empty_to_none=False, cast=True, nan_allowed=False, msg=None):
    def f(value):
        if value in [None, ''] and empty_to_none:
            return None

        try:
            casted_value = float(value)
        except (ValueError, TypeError):
            raise Invalid(msg or 'Given value cannot be casted to float.')

        if not nan_allowed and math.isnan(casted_value):
            raise Invalid(msg or 'Given value is NaN.')

        if cast:
            return casted_value
        return value
    return f


def Decimable(empty_to_none=False, cast=True, nan_allowed=False, msg=None):
    def f(value):
        if value in [None, ''] and empty_to_none:
            return None

        try:
            if isinstance(value, float):
                casted_value = decimal.getcontext()\
                    .create_decimal_from_float(value)
            else:
                casted_value = decimal.Decimal(value)
        except decimal.InvalidOperation:
            raise Invalid(msg or 'Given value cannot be casted to a decimal.')

        if not nan_allowed and math.isnan(casted_value):
            raise Invalid(msg or 'Given value is NaN.')

        if cast:
            return casted_value
        return value
    return f


def Dateable(empty_to_none=False, cast=True, format=None, msg=None):
    if format is None:
        format = '%Y-%m-%d'

    def f(value):
        if value in [None, ''] and empty_to_none:
            return None

        if isinstance(value, datetime.date):
            return value

        try:
            casted_value = datetime.datetime.strptime(value, format)
        except ValueError:
            raise Invalid(msg or 'Given value cannot be casted to a date.')

        if cast:
            return casted_value.date()
        return value
    return f


def Choice(in_list, msg=None):
    def f(value):
        if value not in in_list:
            error_msg = 'Incorrect choice, expected one of the '\
                        'following: "{}".'.format(', '.join(in_list))
            raise Invalid(msg or error_msg)
        return value
    return f


def adapt_dict(input_dict, keep=None, remove=None, make_required=None):
    """Adapt a validation dictionary.

    This is intended to transform "base" dictionaries, with all possible
    arguments for a model, into create and update schemas. The returned
    dict can be used to build a Schema.

    Keyword arguments:
        input_dict (mandatory) -- initial dictionary
        keep -- list of keys to keep
        remove -- list of keys to remove
        make_required -- list of keys to wrap in Required() and keep

    If only `d` is given, return Schema(d).

    If both `keep` and `remove` are provided, `remove` is ignored.
    `make_required` (if provided) is applied over the remaining keys
    after keep or remove, i.e. if a key is removed or not kept and still
    mentioned in the `make_required` list, it will be ignored.

    """
    if keep:
        output_dict = {}
        for k in keep:
            output_dict[k] = input_dict[k]

    elif remove:
        output_dict = input_dict.copy()
        for k in remove:
            output_dict.pop(k)

    else:
        output_dict = input_dict.copy()

    if make_required:
        for k in make_required:
            output_dict[Required(k)] = output_dict.pop(k)

    return output_dict


class SchemaDictNone(Schema):
    """Custom implementation of a dict schema where all values except
    thoses specified in `not_none` (and thoses required) could be None.

    The following are equivalent:
        Schema({
            Required('id'): int,
            'name': unicode,
            'value': Any(None, int),
            'target': Any(None, str),
        })

        SchemaDictNone({
            Required('id'): int,
            'name': unicode,
            'value': int,
            'target': int,
        }, not_none=['name'])

    """

    def __init__(self, schema, required=False, extra=False, not_none=False):
        if not isinstance(schema, dict):
            raise ValueError('This special Schema is intented to be used with '
                             'dict only.')
        Schema.__init__(self, schema, required, extra)
        self._not_none = not_none if not_none is not False else ()

    def __call__(self, data):
        _data = data.copy()
        popped = []

        for k, v in data.items():
            if v is None and k not in self._not_none:
                _data.pop(k)
                popped.append((k, v))

        schema_out = Schema.__call__(self, _data)
        for k, v in popped:
            schema_out[k] = v

        return schema_out



import os

from setuptools import setup, find_packages


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as r_file:
    readme = r_file.read()


setup(
    name='yesaide',
    version='1.1.1',
    license='MIT',
    author='Bastien Gandouet',
    author_email='bastien@ouihelp.fr',
    long_description=readme,
    packages=find_packages(),
    test_suite='tests',
    install_requires=[
        'SQLAlchemy>=1.0',
        'voluptuous>=0.8',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
